(* Copyright 2008, 2009 Michael Brunner *)

(* This file is part of LIPS.

  LIPS is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as 
	published by the Free Software Foundation, either version 3 of 
	the License, or (at your option) any later version.

  LIPS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
	License along with LIPS.  If not, see 
	<http://www.gnu.org/licenses/>. *)

(** This module provides all necessary types and functions to 
	work with the CoreML language. The CoreML languge extends pure
	lambda-calculus with let-bindings (used for abbreviations) and 
	the conditional if-then-else-construct (together with the 
	boolean constants true and false). *)



(** Type for CoreML-variables. Variables are represented by l-strings, the integer was introduced for renaming-purposes. *)
type var = Strng.t * int

(** Type for CoreML-constants. For now only the boolean constants True and False are supported. *)
type cons = 
	| True (** boolean constant true *)
	| False (** boolean constant false *)

(** Type for CoreML-expressions. *)
type t =
    Var of var (** variable *)
  | Cons of cons (** constant *)
  | Abs of var * t (** abstraction, same as abstraction in lambda-calculus *)
  | App of t * t (** application, same as application in lambda-calculus *)
  | Let of var * t * t (** let-binding, the variable should be replaced with the abbreviation inside the body-term, i.e. Let (v, t, b) = App (Abs (v, b), t) *)
  | If of t * t * t (** if-clause containing condition, then-branch and else-branch *)

(** Type used to relate to certain subparts of a CoreML-expression. E.g., the Left subpart of an application App (t, s) relates to t. *)
type dir = 
	| Left 
	| Middle 
	| Right

(** Type for subterm-positons. *)
type pos = dir list



(** CONSTRUCTOR-FUNCTIONS *)

(** Variable-constructor, takes a variable-name as input and returns the corresponding CoreML-expression. *)
val var : Strng.t -> t

(** Abstraction-constructor, takes a list of variable-names and a CoreML-expression as input and returns the corresponding abstraction. *)
val abs : Strng.t list -> t -> t

(** Application-constructor, takes a list of CoreML-expression as input and returns the corresponding application. *)
val app : t list -> t

(** Let-binding-constructor, takes a variable-name, the abbreviation and the body-term and returns the corresponding let-binding. *)
val leti : Strng.t -> t -> t -> t

(** If-clause-constructor, takes the condition, the then-branch and the else-branch as arguments and return sthe corresponding if-clause. *)
val ite : t -> t -> t -> t

(** Constant-constructor, takes a constant as argument and returns the corrspinding CoreML-expression *)
val cons : cons -> t



(** BASIC-FUNCTIONS *)

(** Returns a set containing all variables of a given CoreML-expression. *)
val vars : t -> var St.t

(** Returns the set containing all free variables of a given CoreML-expression. A variable is considered to be free if it is not bound by an abstraction or a let-binding. *)
val fvars : t -> var St.t



(** Substitutes CoreML-expression for variable inside a given CoreML-expression. This function avoids variable-capture by introducing fresh variables when needed. *)
val substitute : var -> t -> t -> t

(** Applies beta-step to a given CoreML-expression. If the expression is not a redex, this function fails with an exception. *)
val beta : t -> t

(** Checks if given CoreML-expression is a redex. *)
val is_redex : t -> bool

(** Retrieves a list of a CoreML-expression's redex-positions. *)
val redexpos : t -> pos list



(** Checks if a given CoreML-expression is in weak-head-normal-form. *)
val is_whnf : t -> bool

(** Checks if a given CoreML-expression is in normal form. *)
val is_nf : t -> bool

(** Checks if a given CoreML-expression is closed, i.e. does not contain any free variables. *)
val is_closed : t -> bool



(** Takes a base-term and a search-term as arguments and retrieves the positions of search-term inside the base-term. *)
val pos : t -> t -> pos list

(** Returns a base-term and a position as arguments and returns the subterm at the given position. This function fails with an exception if the supplied position is not valid for the given base-term. *)
val get : t -> pos -> t



(** Takes a base-term, a replace-term and a position as arguments. This function inserts the replace-term at the given position of the base-term and returns the resulting CoreML-expression. If the supplied position is not valid for the base-term, this function fails with an exception. *)
val replace : t -> t -> pos -> t

(** Takes a base-term and a list of replacements - a replacement is a position-CoreML-expression-pair - as arguments and performs all replacements on base-term. If any position is not valid this function fails with an exception. *)
val replace_all : t -> (pos * t) list -> t



(** REDUCTION-FUNCTIONS *)

(** This function reduces all variable indexs of a given CoreML-expression to a minimum. This operation does not change the meaning of the expression. *)
val reduce_vars : t -> t

(** All reduction-strategies are managed inside this module. A reduction-strategy is a function taking a CoreML-expression as argument and returning a list of replacements. A replacement is a position-CoreML-expression-pair. *)
module Strategies :
  sig
		(** Call-by-value always reduces the (leftmost) innermost redex. This strategy reduces to normal form. *) 
    val call_by_value : t -> (pos * t) list
    
		(** Call-by-name always reduces the (leftmost) outermost redex. This strategy reduces to normal form. *)
		val call_by_name : t -> (pos * t) list
    
		(** Parallel innermost always reduces all innermost redexes in a parallel fashion. This strategy reduces to normal form. *)
		val parallel_inner : t -> (pos * t) list
    
		(** Parallel outermost always reduces all outermost redexes in a parallel fashion. This strategy reduces to normal form. *)
		val parallel_outer : t -> (pos * t) list
  end
	
(** Takes a reduction-strategy and a CoreML-expression as arguments and applies one single beta-reduction. *)	
val do_step : (t -> (pos * t) list) -> t -> t

(** Takes a reduction-strategy and a CoreML-expression as arguments. This function returns the full reduction-sequence in a list. *)
val calc_steps : (t -> (pos * t) list) -> t -> (t * pos list) list

(** Takes a reduction-strategy and a CoreML-expression as arguments and returns the calculated normal form of the given CoreML-expression. *)
val calc_nf : (t -> (pos * t) list) -> t -> t



(** PARSERS *)

(** Combinator Parser for variable names. *)
val varname : Strng.t -> char list Parser.result

(** Combinator Parser for CoreML-expressions. *)
val ltp : t Parser.t



(** PRINTER-FUNCTIONS *)

(** String-converter for variables. Returns the l-string for a given variable-name and index. *)
val var2strng : char list * int -> char list

(** Basic string-converter for CoreML-expressions. Returns the l-string for a given CoreML-expression without using conventions (no pretty printing). *)
val to_strng' : t -> Strng.t

(** String-converter for CoreML-expressions. Returns the l-string for a given CoreML-expression using conventions (pretty printing: nested abstractions, fewer parenthesis). *)
val to_strng : t -> Strng.t

(** Basic string-converter for CoreML-expressions supporting highlighted subterms. Returns the l-string for a given CoreML-expression where all subterms referenced by the provided position-list are highlighted. Does not use conventions for pretty-printing. *)
val to_strng_mark' : pos list -> t -> Strng.t

(** String-converter for CoreML-expressions supporting highlighted subterms. Returns the l-string for a given CoreML-expression where all subterms referenced by the provided position-list are highlighted. Uses conventions for pretty-printing. *)
val to_strng_mark : pos list -> t -> Strng.t
