(* Copyright 2008, 2009 Michael Brunner *)

(* This file is part of LIPS.

  LIPS is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as 
	published by the Free Software Foundation, either version 3 of 
	the License, or (at your option) any later version.

  LIPS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
	License along with LIPS.  If not, see 
	<http://www.gnu.org/licenses/>. *)

type setting = {
	strategy : Strng.t * (CoreML.t -> (CoreML.pos * CoreML.t) list);
	trace_steps: bool;
	simple_output: bool;
	typing: bool;
	verbose: bool;
	record: bool;
	template: Session.template;
	};;
type vars = (Strng.t * CoreML.t) St.t;;
type types = (CoreML.t * Type.t) St.t;;
type state = (setting * vars * types * Session.t);;

let init_setting = {
	strategy= (Strng.of_string "call-by-name", CoreML.Strategies.call_by_name); 
	trace_steps = false; 
	simple_output = true;
	typing = false;
	verbose = true;
	record = false;
	template = Session.default};;
let init_vars = St.empty;;
let init_types = Type.basic_env;;
let init_session = [];;
let init_state = (init_setting, init_vars, init_types, init_session);;

module StateIOM = MonadT.StateT (struct type t = state let init = init_state end) (Monad.IO);;
module M = MonadT.ExceptionT (StateIOM);;
include M;;


let write s = M.lift (StateIOM.lift (Monad.IO.write s));;
let write_line s = M.lift (StateIOM.lift (Monad.IO.write_line s));;
let read_line = M.lift (StateIOM.lift (Monad.IO.read_line));;
let file_exists f = M.lift (StateIOM.lift (Monad.IO.file_exists f));;
let read_file f = M.lift (StateIOM.lift (Monad.IO.read_file f));;
let write_file f contents = M.lift (StateIOM.lift (Monad.IO.write_file f contents));;

let read_state = M.lift (StateIOM.read_state);;
let write_state s = M.lift (StateIOM.write_state s);;

let set_verbose x = 
	let change_verbose old = {strategy = old.strategy; trace_steps = old.trace_steps; simple_output = old.simple_output; typing = old.typing; verbose = x; record = old.record; template = old.template} in 
	read_state >>= fun (s, v, t, u) -> write_state (change_verbose s, v, t, u);;
let get_verbose = 
	read_state >>= fun (s, _, _, _) -> return s.verbose;;
let set_strategy name strategy = 
	let change_strategy old = {strategy = (name, strategy); trace_steps = old.trace_steps; simple_output = old.simple_output; typing = old.typing; verbose = old.verbose; record = old.record; template = old.template} in
	read_state >>= fun (s, v, t, u) -> write_state (change_strategy s, v, t, u) >> 
	get_verbose >>= fun x -> if x then write_line ("Evaluation-strategy changed to " ^ (Strng.to_string name)) else return ();;
let get_strategy_name = 
	read_state >>= fun (s, _, _, _) -> return (fst s.strategy);;
let get_strategy_function = 
	read_state >>= fun (s, _, _, _) -> return (snd s.strategy);;
let set_trace_steps trace = 
	let change_trace_steps old = {strategy = old.strategy; trace_steps = trace; simple_output = old.simple_output; typing = old.typing; verbose = old.verbose; record = old.record; template = old.template} in 
	read_state >>= fun (s, v, t, u) -> write_state (change_trace_steps s, v, t, u) >>
	get_verbose >>= fun x -> if x then write_line ("Tracing is now " ^ (if trace then "on" else "off")) else return ();;
let get_trace_steps = 
	read_state >>= fun (s, _, _, _) -> return s.trace_steps;;
let set_simple_output simple = 
	let change_simple_output old = {strategy = old.strategy; trace_steps = old.trace_steps; simple_output = simple; typing = old.typing; verbose = old.verbose; record = old.record; template = old.template} in 
	read_state >>= fun (s, v, t, u) -> write_state (change_simple_output s, v, t, u) >>
	(*get_template >>= fun tmpl -> let tmpl' = {tmpl with simple_output = simple} in set_template tmpl' >> *)
	get_verbose >>= fun x -> if x then write_line ("Pretty printing is now " ^ (if simple then "on" else "off")) else return ();;
let get_simple_output = 
	read_state >>= fun (s, _, _, _) -> return s.simple_output;;
let get_coreml_converter = 
	get_simple_output >>= fun simple -> return (if simple then CoreML.to_strng else CoreML.to_strng');;
let get_coreml_mark_converter = 
	get_simple_output >>= fun simple -> return (if simple then CoreML.to_strng_mark else CoreML.to_strng_mark');;
let set_typing typing = 
	let change_typing old = {strategy = old.strategy; trace_steps = old.trace_steps; simple_output = old.simple_output; typing = typing; verbose = old.verbose; record = old.record; template = old.template} in 
	read_state >>= fun (s, v, t, u) -> write_state (change_typing s, v, t, u) >>
	get_verbose >>= fun x -> if x then write_line ("Typing is now " ^ (if typing then "on" else "off")) else return ();;
let get_typing = 
	read_state >>= fun (s, _, _, _) -> return s.typing;;
let set_record record = 
	let change_record old = {strategy = old.strategy; trace_steps = old.trace_steps; simple_output = old.simple_output; typing = old.typing; verbose = old.verbose; record = record; template = old.template} in 
	read_state >>= fun (s, v, t, u) -> write_state (change_record s, v, t, u) >>
	get_verbose >>= fun x -> if x then write_line ("Session recording is now " ^ (if record then "on" else "off")) else return ();;
let get_record = 
	read_state >>= fun (s, _, _, _) -> return s.record;;
let set_template template =
	let change_record old = {strategy = old.strategy; trace_steps = old.trace_steps; simple_output = old.simple_output; typing = old.typing; verbose = old.verbose; record = old.record; template = template} in 
	read_state >>= fun (s, v, t, u) -> write_state (change_record s, v, t, u);;
let get_template = 
	read_state >>= fun (s, _, _, _) -> return s.template;;

let has_type t = 
	read_state >>= fun (_, _, types, _) -> return (St.has_one (fun x -> fst x = t) types);;
let get_type t = 
	has_type t >>= function false -> raise "type_not_found" | true ->
	read_state >>= fun (_, _, types, _) -> return (snd (St.find_by (fun x -> fst x = t) types));;
let get_types = 
	read_state >>= fun (_, _, types, _) -> return types;;
let add_type n t = 
	read_state >>= fun (s, vars, types, u) -> write_state (s, vars, St.replace_by (fun x -> fst x = n) (n, Type.reduce_var_index t) types, u) >>
	get_verbose >>= fun x -> if x then write_line ("Type for \"" ^ (Strng.to_string (CoreML.to_strng n)) ^ "\" added to environment") else return ();;
let del_type t = 
	read_state >>= fun (s, vars, types, u) -> write_state (s, vars, St.remove_by (fun x -> fst x = t) types, u);;

let has_var v = 
	read_state >>= fun (_, vars, _, _) -> return (St.has_one (fun x -> fst x = v) vars);;
let get_var v = 
	has_var v >>= function false -> raise "var_not_found" | true ->
	read_state >>= fun (_, vars, _, _) -> return (snd (St.find_by (fun x -> fst x = v) vars));;
let get_vars = 
	read_state >>= fun (_, vars, _, _) -> return vars;;
let add_var n t = 
	read_state >>= fun (s, vars, types, u) -> write_state (s, St.replace_by (fun x -> fst x = n) (n, t) vars, types, u) >>
	get_verbose >>= fun x -> (if x then write_line ("Variable \"" ^ (Strng.to_string n) ^ "\" added to environment") else return ()) >>
	if Type.is_typable types t then let typ = Type.infer_type types t in set_verbose false >> add_type (CoreML.var n) typ >> set_verbose x else return ();;
let del_var v = 
	read_state >>= fun (s, vars, types, u) -> write_state (s, St.remove_by (fun x -> fst x = v) vars, types, u);;

let get_session = 
	read_state >>= fun (_, _, _, session) -> return (Lst.rev session);;
let reset_session = 
	read_state >>= fun (s, vars, types, _) -> write_state (s, vars, types, []);;
let session_add input result = 
	read_state >>= fun(s, vars, types, session) -> write_state (s, vars, types, (Session.Nfe (input, [], result)) :: session);;
let session_add_trace input trace result = 
	read_state >>= fun(s, vars, types, session) -> write_state (s, vars, types, (Session.Nfe (input, Lst.rev (Lst.tl (Lst.rev trace)), result)) :: session);;
let session_addt input result typ = 
	read_state >>= fun(s, vars, types, session) -> write_state (s, vars, types, (Session.TNfe (input, [], result, typ)) :: session);;
let session_addt_trace input trace result typ = 
	read_state >>= fun(s, vars, types, session) -> write_state (s, vars, types, (Session.TNfe (input, Lst.rev (Lst.tl (Lst.rev trace)), result, typ)) :: session);;

let (>?>) = trywith;;
