(* Copyright 2008, 2009 Michael Brunner *)

(* This file is part of LIPS.

  LIPS is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as 
	published by the Free Software Foundation, either version 3 of 
	the License, or (at your option) any later version.

  LIPS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
	License along with LIPS.  If not, see 
	<http://www.gnu.org/licenses/>. *)
	
type t = Quit | Help | Defs | ByName | ByValue | ParallelInner | ParallelOuter | Trace | Nfe of CoreML.t | Cte of CoreML.t * Type.t | Var of Strng.t * CoreML.t | TVar of Strng.t * CoreML.t * Type.t | Run of Strng.t | Load of Strng.t | Typing | Types | Type of CoreML.t * Type.t | Pretty | Record | Flush of Strng.t | Template of Strng.t | Unknown;;


open Parser;;
(** parser for interpreter-commands *)
let cmdp term = 
	(
		(string "!" >> ((string "quit" >> skip_many any_char >> return Quit) 
		            <|> (string "help" >> skip_many any_char >> return Help) 
								<|> (tempt (string "by_name" >> skip_many any_char >> return ByName)) 
								<|> (string "by_value" >> skip_many any_char >> return ByValue)
								<|> (tempt(string "parallel_inner" >> skip_many any_char >> return ParallelInner))
								<|> (tempt(string "parallel_outer" >> skip_many any_char >> return ParallelOuter))
								<|> (tempt (string "trace" >> skip_many any_char >> return Trace))
								<|> (tempt (string "typing" >> skip_many any_char >> return Typing))
								<|> (tempt (string "run " >> many1 any_char >>= fun filename -> skip_many any_char >> return (Run filename))) 
								<|> (tempt (string "load " >> many1 any_char >>= fun filename -> skip_many any_char >> return (Load filename))
							  <|> (tempt (string "types" >> skip_many any_char >> return Types)) 
							  <|> (tempt (string "type " >> CoreML.varname >>= fun v -> spaces >> char ':' >> spaces >> Type.typep >>= fun t -> return (Type (CoreML.var v, t)))) 
							 	<|> (tempt (string "defs" >> skip_many any_char >> return Defs)) 
								<|> (tempt (string "def " >> CoreML.varname >>= fun v -> spaces >> char ':' >> spaces >> Type.typep >>= fun t -> spaces >> char '=' >> spaces >> CoreML.ltp >>= fun l -> many (char ' ') >> return (TVar (v,l, t))))
								<|> (tempt (string "def " >> CoreML.varname >>= fun v -> spaces >> char '=' >> spaces >> CoreML.ltp >>= fun l -> many (char ' ') >> return (Var (v,l)))))
							  <|> (tempt (string "pretty" >> skip_many any_char >> return Pretty)) 
							  <|> (tempt (string "record" >> skip_many any_char >> return Record)) 
							  <|> (tempt (string "flush" >> spaces >> many1 any_char >>= fun filename -> return (Flush filename))) 
							  <|> (tempt (string "template" >> spaces >> many1 any_char >>= fun filename -> return (Template filename))) 
								<|> ((skip_many any_char) >> return Unknown)))
		<|> (tempt (CoreML.ltp >>= fun l -> string ":" >> spaces >> Type.typep >>= fun t -> return (Cte (l, t))))
		<|> (CoreML.ltp >>= fun l -> return (Nfe l))
		<|> (string "\004" >> return Quit)
	) term;;

open InterpreterM;;

(* EVALUATION-FUNCTIONS *)
(** replace all environment-variables within lambda-term (only free variables touched) *)
let rec replace_vars = function
	| CoreML.Var (v, _) as t -> has_var v >>= (function true -> get_var v >>= replace_vars | false -> return t)
	| CoreML.Abs (v, t) -> 
		has_var (fst v) >>= (function 
			| true -> get_verbose >>= fun verbose -> set_verbose false >> get_var (fst v) >>= fun mem -> del_var (fst v) >> replace_vars t >>= fun x -> add_var (fst v) mem >> set_verbose verbose >> return (CoreML.Abs (v, x))
			| false -> replace_vars t >>= fun x -> return (CoreML.Abs (v, x))) 
	| CoreML.App (t1, t2) -> replace_vars t1 >>= fun x -> replace_vars t2 >>= fun y -> return (CoreML.App (x, y))
	| CoreML.Let (v, t, b) -> 
		has_var (fst v) >>= (function 
			| true -> get_verbose >>= fun verbose -> set_verbose false >> get_var (fst v) >>= fun mem -> del_var (fst v) >> replace_vars t >>= fun x -> replace_vars b >>= fun y -> add_var (fst v) mem >> set_verbose verbose >> return (CoreML.Let (v, x, y))
			| false -> replace_vars t >>= fun x -> replace_vars b >>= fun y -> return (CoreML.Let (v, x, y)))
	| CoreML.If (c, t1, t2) -> replace_vars c >>= fun x -> replace_vars t1 >>= fun y -> replace_vars t2 >>= fun z -> return (CoreML.If (x, y, z))
	| CoreML.Cons _ as c -> return c;;


(** print defined named lambda-terms *)
let defs = 
	write_line "Defined named lambda-terms: " >>
	get_vars >>= fun vars ->
	get_coreml_converter >>= fun coreml_to_strng ->
	let rec print_vars = function 
		| [] -> return ()
		| (name, term)::vars -> write_line ((Strng.to_string name) ^ " = " ^ (Strng.to_string (coreml_to_strng term))) >> return vars >>= print_vars
  in print_vars (St.to_list vars);;

(** print defined types *)
let types = 
	write_line "Defined types: " >>
	get_types >>= fun types ->
	get_coreml_converter >>= fun coreml_to_strng ->
	let rec print_types = function 
		| [] -> return ()
		| (term, t)::types -> write_line ((Strng.to_string (coreml_to_strng term)) ^ " : " ^ (Strng.to_string (Type.to_strng t))) >> return types >>= print_types
  in print_types (St.to_list types);;


let check_closed l = 
	if CoreML.is_closed l then 
		return true
	else
		get_verbose >>= fun verbose ->
		(if verbose then 
			write_line "Warning: given term is not closed"
		else 
			return ()) >> 
		return false;;

let check_typable l = 
	get_types >>= fun env ->
	if Type.is_typable env l then 
		return true
	else 
		get_verbose >>= fun verbose ->
		(if verbose then 
			write_line "Error: given term is not well typed" 
		else 
			return ()) >>
		return false;;

(** print execution-trace *)
let print_trace term = 
	get_coreml_mark_converter >>= fun coreml_to_strng_mark ->
	let rec print_trace = function 
		| [] -> return ()
		| (a, b)::xs -> write_line ("Trace: " ^ (Strng.to_string (coreml_to_strng_mark b a))) >> print_trace xs in
	get_trace_steps >>= fun trace ->
	if trace then 
  	get_strategy_function >>= fun strat -> 
		print_trace (CoreML.calc_steps strat term)
	else 
		return ();;

(** write current normal form calculation to session *)
let write_session_nfe input term result = 
	get_record >>= fun record ->
	if record then 
	 	get_strategy_function >>= fun strat -> 
		get_trace_steps >>= fun trace ->
		if trace then 
			session_add_trace input (CoreML.calc_steps strat term) result
		else 
			session_add input result
	else return ();;
(** write current normal form calculation including type-inference result to session *)
let write_session_tnfe input term result tresult = 
	get_record >>= fun record ->
	if record then 
	 	get_strategy_function >>= fun strat -> 
		get_trace_steps >>= fun trace ->
		if trace then 
			session_addt_trace input (CoreML.calc_steps strat term) result tresult
		else 
			session_addt input result tresult
	else return ();;


(** calculate and print normal form *)
let calcnf l = 
	get_strategy_function >>= fun strat -> 
	get_coreml_converter >>= fun coreml_to_strng ->
	get_verbose >>= fun verbose ->
	get_typing >>= fun typing ->
	replace_vars l >>= fun term ->
	check_closed term >>
		if typing then 
			get_types >>= fun env ->
			check_typable l >>= fun ok ->
			if ok then 
				print_trace term >>
  			let result = CoreML.reduce_vars (CoreML.calc_nf strat term) in 
				let tresult = Type.infer_type env l in
				if verbose then 
					write ("NF: " ^ (Strng.to_string (coreml_to_strng result))) >>
					write_line (" : " ^ (Strng.to_string (Type.to_strng (tresult)))) >>
			  	write_session_tnfe l term result tresult
				else 
			  	write_session_tnfe l term result tresult
			else 
				return ()
		else 
			print_trace term >>
			let result = CoreML.reduce_vars (CoreML.calc_nf strat term) in 
			if verbose then 
			  write_line ("NF: " ^ (Strng.to_string (coreml_to_strng result))) >>
			  write_session_nfe l term result
			else 
				write_session_nfe l term result;;

(** check type for term *)
let check_type l t = 
	get_types >>= fun env -> 
	if Type.check_type env l t then 
		write_line ("Type-check: o.k. (has type " ^ (Strng.to_string (Type.to_strng t)) ^ ")" )
	else 
		if Type.is_typable env l then 
			write_line ("Type-check: error (has type " ^ (Strng.to_string (Type.to_strng (Type.infer_type env l))) ^ " and not " ^ (Strng.to_string (Type.to_strng t)) ^ ")") 
		else 
			write_line ("Type-check: error (is not typable)");;

(** flush current session-content to file *)
let flush_session filename = 
	get_session >>= fun session ->
	get_template >>= fun template ->
	let output = Session.to_strng template session in 
	write_file filename output >>= fun ok ->
	if ok then 
		write_line ("Session flushed to \"" ^ filename ^ "\".") >>
		reset_session
	else 
	  return ();;

(** load template from file *)		
let load_template filename = 
	get_template >>= fun template ->
	file_exists filename >>= fun exists ->
	if not exists then 
		write_line "Error opening template-file"
	else 
		read_file filename >>= fun contents -> 
		let contents' = Strng.of_string (Lst.fold1 (fun x xs -> x ^ "\n" ^ xs) contents) in
		let result = Parser.parse (Session.tmplp template) contents' in 
		match result with 
			| Some (template') -> set_template template' >> write_line "Template successfully loaded"
			| _ -> write_line "Error loading template";;

(* INTERPRETER-LOOP-FUNCTIONS *)
(** process single command given as string *)
let rec process_command str = 
	if String.length str = 0 then 
		return false
	else 
		match Parser.parse_strict cmdp (Strng.of_string str) with 
			| Some Quit -> return true
			| Some Unknown -> write_line "Error: unknown command" >> return false
			| Some Help -> Help.msg >> return false
			| Some Defs -> defs >> return false
			| Some Types -> types >> return false
			| Some ByName -> set_strategy (Strng.of_string "call-by-name") CoreML.Strategies.call_by_name >> return false
			| Some ByValue -> set_strategy (Strng.of_string "call-by-value") CoreML.Strategies.call_by_value >> return false
			| Some ParallelInner -> set_strategy (Strng.of_string "parallel innermost") CoreML.Strategies.parallel_inner >> return false
			| Some ParallelOuter -> set_strategy (Strng.of_string "parallel outermost") CoreML.Strategies.parallel_outer >> return false
			| Some Trace -> get_trace_steps >>= fun x -> set_trace_steps (not x) >> return false
			| Some Typing -> get_typing >>= fun x -> set_typing (not x) >> return false
			| Some (Run filename) -> get_verbose >>= fun verbose -> set_verbose false >> process_file (Strng.to_string filename) >>= fun exit -> set_verbose verbose >> return exit
			| Some (Load filename) -> process_file (Strng.to_string filename)
			| Some (Nfe l) -> calcnf l >> return false
			| Some (Cte (l, t)) -> check_type l t >> return false
			| Some (Var (v, l)) -> add_var v l >> return false
			| Some (TVar (v, l, t)) -> add_var v l >> add_type (CoreML.var v) t >> return false
			| Some (Type (v, t)) -> add_type v t >> return false
			| Some (Pretty) -> get_simple_output >>= fun s -> set_simple_output (not s) >> return false
			| Some (Record) -> get_record >>= fun x -> set_record (not x) >> return false
			| Some (Flush filename) -> flush_session (Strng.to_string filename) >> return false
			| Some (Template filename) -> load_template (Strng.to_string filename) >> return false
			| _ -> write_line "syntax error\n" >> return false

(** process given file *)
and process_file filename = 
	get_verbose >>= fun verbose ->
  file_exists filename >>= fun exists ->
	let rec process_lines = function 
		| [] -> return false 
		| x::xs -> (if verbose then write_line (filename ^ "> " ^ x) else return ()) >> 
		  process_command x >>= fun exit ->
			if exit then return true else process_lines xs in
	(if verbose then return () else write ("loading " ^ filename ^ " ...")) >>
	if exists then 
		(read_file filename >>= fun content -> 
		process_lines content >>= fun exit -> 
		if verbose then return exit else write_line ("done (" ^ (string_of_int (Lst.length content)) ^ " lines)") >> return exit)
	else 
		write_line " not found" >> return false;;
	
	

(** interpreter-loop: get input and process it *)
let toplevel = 
	let rec toplevel = fun () ->
		write "> " >> 
		read_line >>= fun input -> 
		process_command input >>= fun quit -> 
		if quit then 
			return () 
		else 
			toplevel ()
	in toplevel ();;

(** print welcome-text *)
let welcome = 
	write_line "Welcome to LIPS - The Lambda-calculus Interpreter for Pure Students" >>
  write_line "Version 1.0.3, 20.02.2009" >>
  write_line "developed by Michael Brunner.\n" >>
	write_line "Type \"!help\" for further information, \"!quit\" to exit.\n";;

(** print goodbye-text *)
let quit = write_line "exiting";;

(** initialize interpreter from default-config-file *)
let init = 
	get_verbose >>= fun verbose -> 
	set_verbose false >>
	process_file ".lambdainit" >>
	set_verbose verbose;;

(* START THE INTERPRETER *)

run (welcome >> init >> toplevel >> quit);;
